from pathlib import Path
from qcodes import initialise_or_create_database_at, config, load_by_run_spec
import numpy as np
import scipy.linalg
import json


def db_to_xarray(ind,add_params_to_snapshot=True, **kwargs):

    """
    Take a dataset from a qcodes database identified by its ID and transform it into a xarray.Dataset
    Wraps around the function load_by_run_spec, which allows to get data from different databases, if you supply
    the corresponding connection. The snapshot of the dataset is available via the attrs attribute of the returned
    xarray.Dataset.

    @param ind: index of the dataset in the QCoDeS database you want to transform to a XArray
    @param kwargs: kwargs to be passed to the underlying qcodes function load_by_run_spec
    @return: xarray.Dataset with the independent parameters as coordinates,
     and the dependent parameters as Data variables
    """

    d = load_by_run_spec(captured_run_id=ind, **kwargs)
    ds = d.to_xarray_dataset()
    if add_params_to_snapshot:
        add_snapshot_params(ds)
    return ds


def add_info_ds(ds, **kwargs):
    """
    for backwardscompatibility made this function call the newer 
    more complete add_snapshot_params function
    """
    return add_snapshot_params(ds, **kwargs)
    
    return ds



def add_vna_params(ds, verbose=False):
    """
    Add parameters to a dataset
    """
    if len(ds.attrs['snapshot']) != 0 :
        snap = json.loads(ds.snapshot)
    else:
        snap = ds.snapshot_dict
        
        
    s21_snap = snap['station']['instruments']['vna']['submodules']['channels']['channels']['vna_S21']['parameters']
    for p in s21_snap.keys():
        try:
            ds.attrs['vna_'+p] = s21_snap[p]['value']
        except KeyError:
            if verbose:
                print("Skipping", p)
    ds.attrs['snapshot'] = ''
    ds.attrs['snapshot_dict'] = snap
    
def add_instr_params(ds, instrument_name = 'qubsrc', verbose=False):
    """
    Add parameters to a dataset
    """
    if len(ds.attrs['snapshot']) != 0 :
        snap = json.loads(ds.snapshot)
    else:
        snap = ds.snapshot_dict
 
    instr_snap = snap['station']['instruments'][instrument_name]['parameters']
    for p in instr_snap.keys():
        try:
            ds.attrs[instrument_name+'_'+p] = instr_snap[p]['value']
        except KeyError:
            if verbose:
                print("Skipping", p)
    ds.attrs['snapshot'] = ''
    ds.attrs['snapshot_dict'] = snap

def add_station_params(ds, verbose=False):
    """
    Add parameters to a dataset
    """
    if len(ds.attrs['snapshot']) != 0 :
        snap = json.loads(ds.snapshot)
    else:
        snap = ds.snapshot_dict
 
    instr_snap = snap['station']['parameters']
    for p in instr_snap.keys():
        try:
            ds.attrs['param'+'_'+p] = instr_snap[p]['value']
        except KeyError:
            if verbose:
                print("Skipping", p)
    ds.attrs['snapshot'] = ''
    ds.attrs['snapshot_dict'] = snap


def add_snapshot_params(ds, verbose=False, instrument_list = None):
    """
    Adds snapshot params to the dataset as attributes for easier viewing compared to snapshot opening
    instrument list should be a list of the available instruments you want to add. Defaults to all instruments that it
    can detect in the snapshot. 

    ADds instruments as ds.instrname_paramname, e.g. ds.qubsrc_frequency or ds.attrs['qubsrc_frequency']
    """
    if len(ds.attrs['snapshot']) != 0 :
        snap = json.loads(ds.snapshot)
    else:
        snap = ds.snapshot_dict

    #other instruments that we want to add:
    if instrument_list is None:
        instrument_list = list(snap['station']['instruments'].keys())

    #we always have a vna and this is a special instrument that has channels
    add_vna_params(ds, verbose=verbose)
    #add all items that are directly in the station as parameters
    add_station_params(ds, verbose=verbose)

    for instr in instrument_list:
        try:
            add_instr_params(ds, instr, verbose=verbose)
        except:
            if verbose:
                print('SMA not found')
            pass


def extract_ABS_spec_power_settings(ds, plot=False):
    """
    Extracts the drive power settings from a dataset (for ABS spectroscopy data)
    """
    default_power = ds.qubsrc_power
    max_power = ds.param_adaptive_drive_limit_power
    freqs = np.arange(0,67e9,10e6)
    powers = np.zeros(freqs.shape)+default_power


    for setting in ds.param_adaptive_drive_settings:
        if setting[2] =='MAX':
            powers[np.logical_and(freqs>=setting[0], freqs<=setting[1])] = max_power
        else:
            powers[np.logical_and(freqs>=setting[0], freqs<=setting[1])] = setting[2]+default_power
    if plot:
        plt.figure()
        plt.plot(freqs*1e-9,powers)
        plt.grid()
        plt.xlabel('drive frequency (GHz)')
        plt.ylabel('power (dBM)')
    return freqs, powers