# author: Lukas J. Splitthoff, Miguel Carrera Beloa
# date: 2023-10-30
# purpose: lumped element implementation hamiltonian of the SSH model

import numpy as np
import scipy.linalg as sp
import matplotlib.pyplot as plt

def ind_total(L0, Lv):
    '''
    Returns the total inductance of the resonator.
    '''
    return (L0**(-1) + Lv**(-1))**(-1)

def cap_total(C0, Cw):
    '''
    Returns the total capacitance of the resonator.
    '''
    return C0 + Cw

def omega(C0, L0, Cw, Lv):
    '''
    Gives the resonance frequency of the resonators in the SSH chain.
    '''
    Lt = ind_total(L0, Lv)
    Ct = cap_total(C0, Cw)
    omega = np.sqrt(1/(Lt*Ct))

    return omega

def v_le(C0, L0, Cw, Lv):
    return omega(C0, L0, Cw, Lv)/2 * ind_total(L0, Lv)/Lv

def w_le(C0, L0, Cw, Lv):
    return omega(C0, L0, Cw, Lv)/2 * Cw/cap_total(C0, Cw)

def hamiltonianSSH_le_simple(n, C0, L0, Cw, Lv):
    '''
    Returns the tight binding Hamiltonian of a finite SSH chain with on-site energy epsilon and hopping amplitudes v and w.
    '''

    # construct the Hamiltonian matrix
    ham = np.zeros((2*n, 2*n))

    for i in range(2*n):
        ham[i, i] = omega(C0, L0, Cw, Lv)
        if i == 2*n-1:
            break
        elif i%2 == 0:
            ham[i, i+1] = v_le(C0, L0, Cw, Lv)
            ham[i+1, i] = v_le(C0, L0, Cw, Lv)
        else:
            ham[i, i+1] = w_le(C0, L0, Cw, Lv)
            ham[i+1, i] = w_le(C0, L0, Cw, Lv)

    # diagonalize the Hamiltonian
    eigvals, eigvecs = sp.eigh(ham)

    return eigvals, eigvecs, ham


def hamiltonianSSH_le_coupling(n, C0, L0, Cw, Lv):
    '''
    Returns the tight binding Hamiltonian of a finite SSH chain with on-site energy epsilon and hopping amplitudes v and w.
    '''

    # construct the Hamiltonian matrix
    ham_total = np.zeros((2*n+2, 2*n+2))
    ham = np.zeros((2*n, 2*n))

    for i in range(2*n):
        ham[i, i] = omega(C0, L0, Cw, Lv)
        if i == 2*n-1:
            break
        elif i%2 == 0:
            ham[i, i+1] = v_le(C0, L0, Cw, Lv)
            ham[i+1, i] = v_le(C0, L0, Cw, Lv)
        else:
            ham[i, i+1] = w_le(C0, L0, Cw, Lv)
            ham[i+1, i] = w_le(C0, L0, Cw, Lv)

    ham_total[1:-1, 1:-1] = ham
    ham_total[0, 1] = w_le(C0, L0, Cw, Lv)
    ham_total[1, 0] = w_le(C0, L0, Cw, Lv)  
    ham_total[-1, -2] = w_le(C0, L0, Cw, Lv)
    ham_total[-2, -1] = w_le(C0, L0, Cw, Lv)


    # diagonalize the Hamiltonian
    eigvals, eigvecs = sp.eigh(ham_total)

    return eigvals, eigvecs, ham_total

def hamiltonianSSH_le_Lvdisorder(n, C0, L0, Cw, Lv_list):
    '''
    Returns the tight binding Hamiltonian of a finite SSH chain with on-site energy epsilon and hopping amplitudes v and w.
    '''

    # construct the Hamiltonian matrix
    ham = np.zeros((2*n, 2*n))

    for i in range(2*n):
        Lv = Lv_list[i//2]
        ham[i, i] = omega(C0, L0, Cw, Lv)
        if i == 2*n-1:
            break
        elif i%2 == 0:
            ham[i, i+1] = v_le(C0, L0, Cw, Lv)
            ham[i+1, i] = v_le(C0, L0, Cw, Lv)
        else:
            ham[i, i+1] = np.sqrt(w_le(C0, L0, Cw, Lv_list[i//2]) * w_le(C0, L0, Cw, Lv_list[(i+1)//2]))
            ham[i+1, i] = np.sqrt(w_le(C0, L0, Cw, Lv_list[i//2]) * w_le(C0, L0, Cw, Lv_list[(i+1)//2]))

    # diagonalize the Hamiltonian
    eigvals, eigvecs = sp.eigh(ham)

    return eigvals, eigvecs, ham

def hamiltonianSSH_le_COL0Lvdisorder(n, C0, L0, Cw, Lv_list):
    '''
    Returns the tight binding Hamiltonian of a finite SSH chain with on-site energy epsilon and hopping amplitudes v and w.
    Constant Cw
    '''

    # construct the Hamiltonian matrix
    ham = np.zeros((2*n, 2*n))

    for i in range(2*n):
        Lv = Lv_list[i//2]
        ham[i, i] = omega(C0[i], L0[i], Cw, Lv)
        if i == 2*n-1:
            break
        elif i%2 == 0:
            ham[i, i+1] = v_le(C0[i], L0[i], Cw, Lv)
            ham[i+1, i] = v_le(C0[i], L0[i], Cw, Lv)
        else:
            ham[i, i+1] = np.sqrt(w_le(C0[i], L0[i], Cw, Lv_list[i//2]) * w_le(C0[i+1], L0[i+1], Cw, Lv_list[(i+1)//2]))
            ham[i+1, i] = np.sqrt(w_le(C0[i], L0[i], Cw, Lv_list[i//2]) * w_le(C0[i+1], L0[i+1], Cw, Lv_list[(i+1)//2]))

    # diagonalize the Hamiltonian
    eigvals, eigvecs = sp.eigh(ham)

    return eigvals, eigvecs, ham


def hamiltonianSSH_le_alldisorder(n, C0, L0, Cw, Lv):
    '''
    Returns the tight binding Hamiltonian of a finite SSH chain with on-site energy epsilon and hopping amplitudes v and w.
    '''

    # construct the Hamiltonian matrix
    ham = np.zeros((2*n, 2*n))

    for i in range(2*n):
        Lvi = Lv[i//2]
        Cwi = Cw[(i+1)//2]

        ham[i, i] = omega(C0[i], L0[i], Cwi, Lvi)
        if i == 2*n-1:
            break
        elif i%2 == 0:
            ham[i, i+1] = np.sqrt(v_le(C0[i], L0[i], Cw[(i+1)//2], Lvi) * v_le(C0[i+1], L0[i+1], Cw[(i+2)//2], Lvi))
            ham[i+1, i] = np.sqrt(v_le(C0[i], L0[i], Cw[(i+1)//2], Lvi) * v_le(C0[i+1], L0[i+1], Cw[(i+2)//2], Lvi))
        else:
            ham[i, i+1] = np.sqrt(w_le(C0[i], L0[i], Cwi, Lv[i//2]) * w_le(C0[i+1], L0[i+1], Cwi, Lv[(i+1)//2]))
            ham[i+1, i] = np.sqrt(w_le(C0[i], L0[i], Cwi, Lv[i//2]) * w_le(C0[i+1], L0[i+1], Cwi, Lv[(i+1)//2]))

    # diagonalize the Hamiltonian
    eigvals, eigvecs = sp.eigh(ham)

    return eigvals, eigvecs, ham


def cost_function_equal(params, lorentzian_peaks):
    '''
    Returns the cost function of the input arrays.
    Difference of the sum of the squares of the elements.
    '''
    C0, L0 , Cw, Lv = params
    Cw_in = Cw
    C0_in = C0
    L0_in = L0
    Lv_in = Lv
    eigenvalues = hamiltonianSSH_le_simple(n=5, C0=C0_in, L0=L0_in, Cw=Cw_in, Lv=Lv_in)[0]/(2*np.pi)
    eigenvalues = np.array(eigenvalues)
    # print(eigenvalues/1e9-lorentzian_peaks/1e9)
    return np.sum((eigenvalues/1e9-lorentzian_peaks/1e9)**2)

def cost_function_COL0Lvdisorder(params, lorentzian_peaks):
    '''
    Returns the cost function of the input arrays.
    Difference of the sum of the squares of the elements.
    '''
    n = int(len(lorentzian_peaks)/2)
    # C0, L0 , Cw, Lv = params
    Cw_in = params[4*n]
    C0_in = params[:2*n]
    L0_in = params[2*n:4*n]
    Lv_in = params[4*n+1:]
    eigenvalues = hamiltonianSSH_le_COL0Lvdisorder(n=n, C0=C0_in, L0=L0_in, Cw=Cw_in, Lv_list=Lv_in)[0]/(2*np.pi)
    eigenvalues = np.array(eigenvalues)
    # print(eigenvalues/1e9-lorentzian_peaks/1e9)
    return np.sum((eigenvalues/1e9-lorentzian_peaks/1e9)**2)


def cost_function_alldisorder(params, lorentzian_peaks):
    '''
    Returns the cost function of the input arrays.
    Difference of the sum of the squares of the elements.
    '''
    n = int(len(lorentzian_peaks)/2)
    # C0, L0 , Cw, Lv = params

    C0_in = params[:2*n]
    L0_in = params[2*n:4*n]
    Cw_in = params[4*n:5*n+1]
    Lv_in = params[5*n+1:]
    eigenvalues = hamiltonianSSH_le_alldisorder(n=n, C0=C0_in, L0=L0_in, Cw=Cw_in, Lv=Lv_in)[0]/(2*np.pi)
    eigenvalues = np.array(eigenvalues)
    # print(eigenvalues/1e9-lorentzian_peaks/1e9)
    return np.sum((eigenvalues/1e9-lorentzian_peaks/1e9)**2)